/******************************************************************************
 * program:     rasimg library                                                *
 * function:    Executable for image conversion.                              *
 * modul:       convert.cc                                                    *
 * licency:     GPL or LGPL                                                   *
 ******************************************************************************/
#include <stddef.h>
#include <stdio.h>

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "typedfs.h"
#include "image.h"
#include "imageop.h"
#include "mor_tool.h"
#include "img_tool.h"

#include <matrix.h>


/////////////////////Main CODE starts here/////////////////////////////////

char Verbosity = 0;

void PrintHeader(void)
{
char HdrPrinted=0;
  if(Verbosity>=0 && HdrPrinted==0)
  {
    printf("\n\r <<< GCONVERT >>> Convert picture format (c)1997-2024 F&TSoft, version %u.%u\n", (RAS_IMG_VERSION>>8), (RAS_IMG_VERSION & 0xFF));
    HdrPrinted = 1;
  }
}


Raster2DAbstract *ResizeRaster(Raster2DAbstract *RIn, uint16_t Size1D, uint16_t Size2D)
{
  if(RIn==NULL || RIn->Data2D==NULL || RIn->Size1D==0 || RIn->Size2D==0) return RIn;

  if(RIn->Channels()==3 || RIn->Channels()==4)
  {
    Raster2DAbstract *ROut = ((RIn->Channels()==4) ?
		CreateRaster2DRGBA(Size1D,Size2D,RIn->GetPlanes()/4) : CreateRaster2DRGB(Size1D,Size2D,RIn->GetPlanes()/3));
    if(ROut==NULL) goto AllocationProblem;
    RGBQuad RGB;
    for(unsigned y=0; y<Size2D; y++)
    {
      Raster1DAbstract *ROut1 = ROut->GetRowRaster(y);
      if(ROut1==NULL) continue;
      Raster1DAbstract *RIn1 = RIn->GetRowRaster(((unsigned)RIn->Size2D*y)/Size2D);
      if(RIn1==NULL) continue;

      for(unsigned x=0; x<Size1D; x++)
      {
        RIn1->Get(((unsigned)RIn->Size1D*x)/Size1D,&RGB);
        ROut1->Set(x,&RGB);
      }
    }
    return ROut;
  }

  if(RIn->GetPlanes() < 0)
  {
    Raster2DAbstract *ROut = CreateRaster2D(Size1D,Size2D,RIn->GetPlanes());
    if(ROut==NULL) goto AllocationProblem;
    for(unsigned y=0; y<Size2D; y++)
    {
      Raster1DAbstract *ROut1 = ROut->GetRowRaster(y);
      if(ROut1==NULL) continue;
      Raster1DAbstract *RIn1 = RIn->GetRowRaster(((unsigned)RIn->Size2D*y)/Size2D);
      if(RIn1==NULL) continue;

      for(unsigned x=0; x<Size1D; x++)
      {      
        ROut1->SetValue1Dd(x, RIn1->GetValue1Dd(((unsigned)RIn->Size1D*x)/Size1D));
      }
    }
    return ROut;
  }

  {
    Raster2DAbstract *ROut = CreateRaster2D(Size1D,Size2D,RIn->GetPlanes());
    if(ROut==NULL) goto AllocationProblem;
    for(unsigned y=0; y<Size2D; y++)
    {
      Raster1DAbstract *ROut1 = ROut->GetRowRaster(y);
      if(ROut1==NULL) continue;
      Raster1DAbstract *RIn1 = RIn->GetRowRaster(((unsigned)RIn->Size2D*y)/Size2D);
      if(RIn1==NULL) continue;

      for(unsigned x=0; x<Size1D; x++)
      {      
        ROut1->SetValue1D(x, RIn1->GetValue1D(((unsigned)RIn->Size1D*x)/Size1D));
      }
    }
    return ROut;
  }

AllocationProblem:
  printf("Image allocation error %u x %u\n",Size1D,Size2D);
  return RIn;
}


void PrintHelp(void)
{
TImageFileHandler *LoadPicture_XXX;

  printf("Supported file formats are:\n");
  for(LoadPicture_XXX=TImageFileHandler::first();LoadPicture_XXX!=NULL;LoadPicture_XXX=LoadPicture_XXX->next())
  {
    printf("%s%s%s %s%s; ",LoadPicture_XXX->extension(),
			LoadPicture_XXX->description()==NULL?"":"#",
			LoadPicture_XXX->description()==NULL?"":LoadPicture_XXX->description(),
			LoadPicture_XXX->LoadPicture==NULL?"":"R",
			LoadPicture_XXX->SavePicture==NULL?"":"W");
  }
  printf(//"\n/NoScale      - means truncating image data"
	       "\n/D | /Dilatation - Morphology dilatation"
	       "\n/E | /Erosion - Morphology erosion"
	       "\n/El4 | /El8   - Morphology element"
	       "\n/Format \"str\" - specify output format"
	       "\n/Gray         - discards all color informations"
	       "\n/Invert       - inverts image"
	       "\n/Separate[R|G|B|IDX] - remove one channel from true color images"
	       "\n/PeelPlane n  - cut only plane with number n"
	       "\n/PeelFrame n  - cut only scene with number n"
	       "\n/Resize factor% - resize all rasters with given factor"
	       "\n/s            - decrease verbosity level"
	       "\n/v            - increase verbosity level"
	       "\n");
  printf("Speciffic encoder options:\n"
	       "QUALITY nn	- set quality for JPEG writer.\n"
	       "\n");
}


const TImageFileHandler *FindImgHandler(const char *OutFormat)
{
const TImageFileHandler *SavePicture_XXX;
  for(SavePicture_XXX=TImageFileHandler::first(); SavePicture_XXX!=NULL; SavePicture_XXX=SavePicture_XXX->next())
  {
    if(SavePicture_XXX->SavePicture!=NULL)
    {
      const char *ext = SavePicture_XXX->extension();

      if(ext!=NULL)
      {
        if(*ext=='.') ext++;
	size_t i = strlen(ext);
        if(!stricmp(OutFormat,ext))	// Exact match
        {
          return SavePicture_XXX;
	}
	if(!strnicmp(OutFormat,ext,i))
        {
	  if(OutFormat[i]=='#')
	  {
	    if(!stricmp(OutFormat+i+1,SavePicture_XXX->description()))
	    {
              return SavePicture_XXX;
	    }
	  }
        }
      }
    }
  }
  return NULL;
}


/////////////////////////////////////////////

class AImageFilter
{
public:
  AImageFilter(void);
  virtual ~AImageFilter();
  virtual void ProcessFilter(Image *Im) = 0;

  AImageFilter *Next;
  static AImageFilter *First;
  static void Rewind(void);
};

 AImageFilter *AImageFilter::First = NULL;

AImageFilter::AImageFilter()
{
  if(First==NULL)
  {
    First = this;
    Next = NULL;
  }
  else
  {
    Next=First;
    First=this;
  }
}

AImageFilter::~AImageFilter()
{
  if(First==this)
  {
    First = Next;
    return;
  }
}


void AImageFilter::Rewind(void)
{
  AImageFilter *NewFirst = NULL;
  while(First!=NULL)
  {
    AImageFilter * const Curr = First;
    First = First->Next;
    Curr->Next = NewFirst;
    NewFirst = Curr;
  }
  First = NewFirst;
}

//////////////////////////////////

class GrayImageFilter: AImageFilter
{
public:
  virtual void ProcessFilter(Image *Im);
};

void GrayImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rConverting to gray ");
 while(Im != NULL)
 {
   Convert2Gray(*Im);
   Im->AttachPalette(NULL);
   Im = Im->Next;
 }
}


//////////////////////////////////


class InvertImageFilter: AImageFilter
{
public:
  virtual void ProcessFilter(Image *Im);
};

void InvertImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rInverting ");
 while(Im != NULL)
 {
   if(Im->Palette)
   {
     NotR((char*)Im->Palette->Data1D, (Im->Palette->Size1D*Im->Palette->GetPlanes()+7)/8);
   }
   else
   {
     if(Im->Raster && Im->Raster->GetPlanes()>0)
     {
       for(unsigned y=0; y<Im->Raster->Size2D; y++)
       {
         NotR((char*)Im->Raster->GetRow(y), (Im->Raster->Size1D*Im->Raster->GetPlanes()+7)/8);
       }
     }
   }
   Im = Im->Next;
 }
}


//////////////////////////////////


class FlipImageFilter: AImageFilter
{
public:
  virtual void ProcessFilter(Image *Im);
};

void FlipImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rFlip ");
 while(Im != NULL)
 {
   if(Im->Raster)
     Flip1D(Im->Raster);
   Im = Im->Next;
 }
}


class FlopImageFilter: AImageFilter
{
public:
  virtual void ProcessFilter(Image *Im);
};

void FlopImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rFlip ");
 while(Im != NULL)
 {
   if(Im->Raster)
     Flip1D(Im->Raster);
   Im = Im->Next;
 }
}


//////////////////////////////////

class DilateImageFilter: AImageFilter
{
public:
  DilateImageFilter(uint16_t new_MorphType=El8): MorphType(new_MorphType) {};
  virtual void ProcessFilter(Image *Im);

  uint16_t MorphType;
};


void DilateImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rDilatation ");
 while(Im != NULL)
 {
   Dilatation(Im->Raster, MorphType);
   Im = Im->Next;
 }
}

//////////////////////////////////

class ErodeImageFilter: AImageFilter
{
public:
  ErodeImageFilter(uint16_t new_MorphType=El8): MorphType(new_MorphType) {};
  virtual void ProcessFilter(Image *Im);

  uint16_t MorphType;
};


void ErodeImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rErosion ");
 while(Im != NULL)
 {
   Erosion(Im->Raster, MorphType);
   Im = Im->Next;
 }
}


//////////////////////////////////

class ResizePercentImageFilter: AImageFilter
{
public:
  ResizePercentImageFilter(float NewPercent): Percent(NewPercent) {};
  virtual void ProcessFilter(Image *Im);

  float Percent;
};


void ResizePercentImageFilter::ProcessFilter(Image *Im)
{
 if(Verbosity>=0) printf("\rResize ");
 while(Im != NULL)
 {
   if(Im->Raster->Data2D!=NULL)
   {
     Im->AttachRaster(ResizeRaster(Im->Raster, Im->Raster->Size1D*Percent/100, Im->Raster->Size2D*Percent/100));
   }
   Im = Im->Next;
 }
}


/////////////////////////////////////////////


int main(int argc,char *argv[])
{
Image I1, I2;
const char *InputImg=NULL, *OutputImg=NULL;
const char *s;
const char *OutFormat=NULL;
int i;
int PeelPlaneNo = -1;
int PeelFrameNo = -1;
int Separate = -1;
int Error = 0;
Image *CurrImg = &I1;
uint16_t MorphType = El8;
PropertyList SaveParams;

	/* Parse Arguments */
 for(i=1; i<argc; i++)
 {
   s = argv[i];
   if(s==NULL) continue;
   if(*s==0) continue;   

   if(*s=='-' || *s=='/')
   {
     if(strcmp(s+1,"D")==0 || strcmp(s+1,"Dilatation")==0)
     {
       new DilateImageFilter(MorphType);
       continue;
     }
     if(strcmp(s+1,"E")==0 || strcmp(s+1,"Erosion")==0)
     {
       new ErodeImageFilter(MorphType);
       continue;
     }
     if(strcmp(s+1,"El4") == 0)
     {
       MorphType = El4;
       continue;
     }
     if(strcmp(s+1,"El8") == 0)
     {
       MorphType = El8;
       continue;
     }

     if(strcmp(s+1,"help")==0 || strcmp(s+1,"-help")==0 || strcmp(s+1,"?")==0)
     {
       PrintHeader();
       PrintHelp();
       return(0);
     }
     if(strcmp(s+1,"Format")==0 || strcmp(s+1,"format")==0)
     {
       ++i;
       if(i>=argc) break;
       OutFormat = argv[i];
       continue;
     }
     if(strcmp(s+1,"PeelFrame") == 0)
     {
       ++i;
       if(i>=argc) break;
       PeelFrameNo = atoi(argv[i]);
       if(PeelFrameNo <= 0)
       {
         printf("\n\rInvalid scene number: %d.", PeelFrameNo);
         goto End;
       }
       continue;
     }
     if(strcmp(s+1,"PeelPlane") == 0)
     {       
       ++i;
       if(i>=argc) break;
       PeelPlaneNo = atoi(argv[i]);
       if(PeelPlaneNo <= 0)
       {
         printf("\n\rInvalid plane number: %d.", PeelPlaneNo);
         goto End;
       }       
       continue;
     }
     if(strcmp(s+1,"s") == 0)
     {
       Verbosity = -1;
       continue;
     }
     if(strcmp(s+1,"v") == 0)
     {
       Verbosity = +1;
       continue;
     }
     if(strcmp(s+1,"Flip")==0 || strcmp(s+1,"flip")==0)
     {
       new FlipImageFilter;
       continue;
     }
     if(strcmp(s+1,"Gray")==0 || strcmp(s+1,"gray")==0)
     {
       new GrayImageFilter;
       continue;
     }
     if(strcmp(s+1,"Invert")==0 || strcmp(s+1,"invert")==0)
     {
       new InvertImageFilter;
       continue;
     }
     if(strcmp(s+1,"Resize")==0 || strcmp(s+1,"resize")==0)
     {
       ++i;
       if(i>=argc) break;
       const float ResizeScale = atof(argv[i]);
       if(ResizeScale <= 0)
       {
         printf("\n\rInvalid resize factor: %f.", ResizeScale);
         goto End;
       }
       new ResizePercentImageFilter(ResizeScale);
     }

     if(strcmp(s+1,"SeparateR") == 0)
     {
       Separate = 0;
       continue;
     }
     if(strcmp(s+1,"SeparateG") == 0)
     {
       Separate = 1;
       continue;
     }
     if(strcmp(s+1,"SeparateB") == 0)
     {
       Separate = 2;
       continue;
     }
     if(strcmp(s+1,"SeparateIDX") == 0)
     {
       Separate = 10;
       continue;
     }
     if(strcmp(s+1,"Separate") == 0)
     {
       ++i;
       if(i>=argc) break;
       if(strcmp(s,"R")==0) Separate=0;
       else if(strcmp(s,"G")==0) Separate=1;
       else if(strcmp(s,"B")==0) Separate=2;
       else if(strcmp(s,"IDX")==0) Separate=10;
       else
       {
         printf("\n\rInvalid channel for separation: %s.", s);
         goto End;
       }
       continue;
     }
     if(strcmp(s+1,"JPG_QUALITY")==0 || strcmp(s+1,"QUALITY")==0)
     {
       ++i;
       if(i>=argc) break;
       PropertyItem *Prop = new PropertyItem("QUALITY",sizeof(int));
       if(Prop->Data)
       {
         *(int*)Prop->Data = atoi(argv[i]);
         SaveParams.AddProp(Prop);
       }
       else
         delete(Prop);
       continue;
     }
     if(strcmp(s+1,"COMPRESSION")==0)
     {
       ++i;
       if(i>=argc) break;
       PropertyItem *Prop = new PropertyItem("COMPRESSION",sizeof(int));
       if(Prop->Data)
       {
         *(int*)Prop->Data = atoi(argv[i]);
         SaveParams.AddProp(Prop);
       }
       else
         delete(Prop);
       continue;
     }

     if(strcmp(s+1,"jpeg:optimize-coding")==0 || strcmp(s+1,"jpeg:optimize-coding=1")==0)
     {
       PropertyItem *Prop = new PropertyItem("jpeg:optimize-coding",sizeof(char));
       if(Prop->Data)
       {
         *(char*)Prop->Data = 1;
         SaveParams.AddProp(Prop);
       }
       else
         delete(Prop);
       continue;
     }
   }

   if(InputImg==NULL)
     {InputImg=s;continue;}
   else
     if(OutputImg==NULL) {OutputImg=s;continue;}

   printf("Unknown argument \"%s\"!\n", s);
  }
	/* End of reading arguments */

  PrintHeader();
  if(Verbosity>=0) printf("\rLoading");
  //MarkTime=GetInterval();

//for(i=1;i<=10;i++)
  I1 = LoadPicture(InputImg);

  if(I1.isEmpty())
  {
    printf("\n\rCannot load input picture:\"%s\"",InputImg);
    goto End;
  }

  if(PeelFrameNo > 0)
  {
    while(PeelFrameNo>0 && CurrImg!=NULL)
    {
      if(PeelFrameNo==1) break;

      PeelFrameNo--;

      Image * const PImg = CurrImg->Next;
      CurrImg->Next = NULL;
      delete  CurrImg;
      CurrImg = PImg;
    }
    if(CurrImg==NULL)
    {
      printf("\n\rCannot obtain requested scene!");
      goto End;
    }
    if(CurrImg->Next != NULL)
    {
      delete CurrImg->Next;	// This deletes whole remaining chain.
      CurrImg->Next = NULL;
    }
  }

  if(Separate>=0)
  {
    if(Separate!=10)
      *CurrImg = Peel8bit(*CurrImg, Separate);	// R,G,B must be separated BEFORE palatte is cut.

    CurrImg->AttachPalette(NULL);
  }

	// Apply image filters.
  AImageFilter::Rewind();
  while(AImageFilter::First!=NULL)
  {
    AImageFilter::First->ProcessFilter(CurrImg);
    delete AImageFilter::First;
  }

  if(PeelPlaneNo>0)
  {    
    *CurrImg = PeelPlane(*CurrImg, PeelPlaneNo);
    CurrImg->AttachPalette(NULL);
  }


	// Apply properties for a writer.
  for(i=0; i<SaveParams.PropCount; i++)
    CurrImg->AttachProperty(SaveParams.pProperties[i]);

  if(Verbosity>=0) printf("\rSaving ");
  //MarkTime=GetInterval();

  {
    const TImageFileHandler *SavePicture_XXX;
    if(OutFormat==NULL)
    {				//No format requested
      SavePicture_XXX = FindImgHandler(ReadExt(OutputImg));
    }
    else
    {
      SavePicture_XXX = FindImgHandler(OutFormat);
    }

    if(SavePicture_XXX != NULL)
    {
      if((SavePicture_XXX->Flags&FILE_FORMAT_CAP_EXIF)==0 && HasExif(CurrImg))
         AutoOrient(CurrImg);
      Error = SavePicture_XXX->SavePicture(OutputImg,*CurrImg);
    }
    else
        Error = 0xFFFF;
  }

  if(Error)
  {
    if(OutFormat && Error==0xFFFF)
      printf("\n\rNo export format:\"%s\" found.",OutFormat);
    else
      printf("\n\rCannot save output picture:\"%s\" Error %d",OutputImg,Error);
  }
  else
  {
    if(Verbosity>=0)
    {
      //printf("time : %f",abs(GetInterval()-MarkTime)/18.2);
      puts("\rOK     ");
    }
  }

End:
  return(Error);
}

